#include <stdio.h>
#include <string.h>
#include "e3d.h"

#define TRUE	1
#define FALSE	0

// func from 3DOUtility.c
extern void ResetModel(ThreeDeeModel *pmodel);
extern int CalculateNormals(ThreeDeeModel *pmodel);

// Load in a 3DO model
// note: fix return values
int Load3DO(char *filename, ThreeDeeModel *pmodel) {

	int validcmd, validparam, end, n, i, j, numverts, numnorms, numpolys, vert;
	float x, y, z, r, g, b;
	char scmd[80], stemp[80];
	FILE *pfile;

	fprintf(stderr, "Opening model file %s\n", filename);
	pfile = fopen(filename, "r");
	// check if file open failed
	if(pfile == NULL) {
		fprintf(stderr, "Model file open failed.\n");
        return(5);
	}

	// Reset model values
    ResetModel(pmodel);

	// Parse model file
	end = FALSE;
	numverts = 0;
    numnorms = 0;
	numpolys = 0;
	//modelnum = myscene.nummodels;         !legacy!
	while(!end) {
		// reset valid flags to trap errors
		validcmd = FALSE; validparam = FALSE;

		fscanf(pfile, "%s", &scmd[0]);
		fprintf(stderr, "Command is %s\n", scmd);
		if(scmd[0] == '!') {
			validcmd = TRUE;
			fprintf(stderr, "Comment encountered. Scanning till eol.\n");
			// read till end of line (character 10)
			fscanf(pfile, "%c", &stemp[0]);
			while(stemp[0] != 10) fscanf(pfile, "%c", &stemp[0]);
		}
		if(strcmp(scmd, "vert") == 0) {
        	if(numverts == MAX_VERTS_PER_MODEL - 1 ) {
        		fprintf(stderr, "Error - too many verts! Exiting...\n");
				fclose(pfile);
				return(6);		// error
			}
			validcmd = TRUE;
			fscanf(pfile, "%f %f %f",
				&x, &y, &z);
			fprintf(stderr, "vert: xyz %3.3f %3.3f %3.3f\n",
				x, y, z);
			pmodel->vert[numverts].x = x;
			pmodel->vert[numverts].y = y;
			pmodel->vert[numverts].z = z;
            // verts default colour
			pmodel->vert[numverts].r = 0.5;
			pmodel->vert[numverts].g = 0.5;
			pmodel->vert[numverts].b = 0.5;
			numverts++;
		}

		if(strcmp(scmd, "norm") == 0) {
        	if(numnorms == MAX_VERTS_PER_MODEL - 1 ) {
        		fprintf(stderr, "Error - too many normals! Exiting...\n");
				fclose(pfile);
				return(6);		// error
			}
			validcmd = TRUE;
			fscanf(pfile, "%f %f %f",
				&x, &y, &z);
			fprintf(stderr, "norm: xyz %3.3f %3.3f %3.3f\n",
				x, y, z);
			pmodel->norm[numnorms].x = x;
			pmodel->norm[numnorms].y = y;
			pmodel->norm[numnorms].z = z;
			numnorms++;
		}

		if(strcmp(scmd, "poly") == 0) {
        	if(numpolys == MAX_POLYS_PER_MODEL - 1 ) {
        		fprintf(stderr, "Error - too many polys! Exiting...\n");
				fclose(pfile);
				return(7);		// error
			}
			validcmd = TRUE;
            // set poly default values
			pmodel->poly[numpolys].shade = 0;
			pmodel->poly[numpolys].gouraud = 0;
            pmodel->poly[numpolys].filling = 0;
            pmodel->poly[numpolys].colour.r = 0.5;
            pmodel->poly[numpolys].colour.g = 0.5;
            pmodel->poly[numpolys].colour.b = 0.5;
            // one normal per poly, for now...
            pmodel->poly[numpolys].n[0] = numpolys;
			// get no of verts
			fscanf(pfile, "%d", &n);
			fprintf(stderr, "poly: %d verts   ", n);
			pmodel->poly[numpolys].numverts = n;
			// read in vert indices
			for(i=0; i<n; i++) {
				fscanf(pfile, "%d", &vert);
				fprintf(stderr, "%d ", vert);
				pmodel->poly[numpolys].p[i] = vert-1;
			}
			// read in texture name
			fscanf(pfile, "%s", &stemp[0]);
			fprintf(stderr, "%s\n", stemp);
			// set texture index here
			numpolys++;
		}

        // colour for poly and for verts
		if(strcmp(scmd, "colour") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d %f %f %f",
				&n, &r, &g, &b);
			fprintf(stderr, "poly %d colour %d   %3.3f %3.3f %3.3f\n",
				j, n, r, g, b);
			if(n == 0) {
				pmodel->poly[j].colour.r = r;
				pmodel->poly[j].colour.g = g;
				pmodel->poly[j].colour.b = b;
				pmodel->poly[j].vcol[0].r = r;
				pmodel->poly[j].vcol[0].g = g;
				pmodel->poly[j].vcol[0].b = b;
			}
            else {
            	if(n <= pmodel->poly[j].numverts) {
					pmodel->poly[j].vcol[n-1].r = r;
					pmodel->poly[j].vcol[n-1].g = g;
					pmodel->poly[j].vcol[n-1].b = b;
				}
                else {
			    	fprintf(stderr, "Error - vertex index exceeds poly numverts! Exiting...\n");
					fclose(pfile);
					return(2);		// error
				}
			}
		}

        // texture map ("texture page") index
		if(strcmp(scmd, "texture") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d", &n);
			fprintf(stderr, "poly %d texture %d\n", j, n);
            	if((n > -1) && (n < 16)) {
					pmodel->poly[j].texmap = n;
				}
                else {
			    	fprintf(stderr, "Error - texture map index out of bounds! Exiting...\n");
					fclose(pfile);
					return(2);		// error
				}
		}

        // texture co-ordinates
		if(strcmp(scmd, "texvert") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d %f %f",
				&n, &x, &y);
			fprintf(stderr, "poly %d texvert %d   %3.1f %3.1f\n",
				j, n, x, y);
            	if(n <= pmodel->poly[j].numverts) {
					pmodel->poly[j].texp[n].x = x;
					pmodel->poly[j].texp[n].y = y;
				}
                else {
			    	fprintf(stderr, "Error - texture vertex index exceeds poly numverts! Exiting...\n");
					fclose(pfile);
					return(2);		// error
				}
		}

		if(strcmp(scmd, "filling") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d",
				&n);
			fprintf(stderr, "poly %d filling %d\n",
				j, n);
			pmodel->poly[j].filling = n;
		}

		if(strcmp(scmd, "shade") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d",
				&n);
			fprintf(stderr, "poly %d shade %d\n",
				j, n);
			pmodel->poly[j].shade = n;
		}

		if(strcmp(scmd, "gouraud") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d",
				&n);
			fprintf(stderr, "poly %d gouraud %d\n",
				j, n);
			pmodel->poly[j].gouraud = n;
		}

		if(strcmp(scmd, "end") == 0) {
			validcmd = TRUE;
			fprintf(stderr, "End of model file.\n");
			end = TRUE;
			// set model parameters
			pmodel->numverts = numverts;
            pmodel->numnorms = numnorms;
			pmodel->numpolys = numpolys;
			strcpy(pmodel->name, filename);
			fprintf(stderr, "%d verts and %d polys loaded.\n", numverts, numpolys);
		}

		// check if no valid command found (ie: parse error)
		if(!validcmd) {
			fprintf(stderr, "Error - Model Command not recognised! Exiting...\n");
			fclose(pfile);
			return(2);		// error
		}

  // check if EOF before end
  if(feof(pfile) && !end) {
			fprintf(stderr, "Error - EOF before end marker! Exiting...\n");
			fclose(pfile);
			return(3);		// error
		}

	}

	fclose(pfile);

	CalculateNormals(pmodel);
	return(0);		// success!
}


